--------------------------------------------------------------------------------
-- Copyright(c) 1997 Rational Software Corporation.  All rights reserved.
-- Proprietary and confidential information of Rational.  Use shall be for the
-- sole purpose of creating an executable, any other use, duplication, or
-- disclosure without the express written authorization by Rational is
-- strictly prohibited.  This material is protected as an unpublished work
-- under the U.S.  Copyright Act of 1976.  Created.
--------------------------------------------------------------------------------




-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
--
--   Rational Software Corporation                 COPYRIGHT (C) 1993-1997
--   2800 San Tomas Expressway                     Permission is hereby granted
--   Santa Clara, California 95051                 to licensees to distribute
--                                                 and use this information
--                                                 for the sole purpose of 
--                                                 building and distributing
--                                                 systems using VADS or Apex.
--
--      FILE: v_usr_conf.2.ada
--
--      UNIT: Apex User Program Configuration Package Body for
-- THREADS/SPARC
--
--
--   PURPOSE: This file contains the body of the package V_USR_CONF.
--            This package body must be modified by the user to provide
--            the information needed to configure the Apex run-time
--            environment for a user application program.
--
--            The components of this package are defined and described
--            in the specification of V_USR_CONF, which is
--            contained in the file v_usr_conf.1.ada.  DO NOT ATTEMPT TO 
--            CONFIGURE APEX BEFORE READING THIS SPECIFICATION.
--
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
with Machine_Code;
with System;
use System;
with System.Runtime;
use System.Runtime;
with System.Return_Address;
with V_Usr_Conf_I;
use V_Usr_Conf_I;
with Ada_Krn_I;
with Ada_Krn_Defs;
with V_Ada_Info;
with V_Usr_Data;
with V_Usr_Local;
with V_I_Types;
with V_I_Sig;
with V_I_Cifo;
with Interfaces.C;
with System.Errno_Support;
with V_I_Main;
with Unchecked_Conversion;
with V_Unsigned_Types;
with Link_Block;
with V_I_Libop;
with Os_Alloc;
with Psignal;
with Taskdeb_I;
with System.Storage_Elements;
with Putil;
with Pthread;
with Unix_Types;
with Unix_Operations;


package body V_Usr_Conf is

    pragma Suppress (All_Checks);

------------ SMALL_BLOCK_SIZES_TABLE structure ----------


    subtype Small_Block_Sizes is V_Usr_Conf_I.Small_Block_Sizes_T (1 .. 4);
    Small_Block_Sizes_Table : constant Small_Block_Sizes := (8, 16, 32, 64);

------------ Memory allocation parameters -----

    Mem_Alloc_Conf_Table : constant V_Usr_Conf_I.Mem_Alloc_Conf_Table_T :=
       (Min_Size => 8,
        Num_Small_Block_Sizes => Small_Block_Sizes_Table'Length,
        Small_Block_Sizes_Address => Small_Block_Sizes_Table'Address,
        Min_List_Length => 5,
        Malloc_Based_Allocation => True,
        External_Pool_Protection => None,
        Private_Storage_Pools => False,
        Storage_Pool_Configuration => (others => Use_Standard_Pool_Routine),
        Allocation_Strategy => First_Fit);

------------------ CCA CONFIGURATION DATA ------------------------------


    Cca_Trap : constant V_Unsigned_Types.Unsigned_Integer := 16#91D02090#;
    Cca_Signal : constant Integer := Psignal.Sigill;
    pragma External_Name (Cca_Trap, "_CCA_TRAP");
    pragma External_Name (Cca_Signal, "_CCA_SIGNAL");

--
-- For information on Unix profiling see prof(1) and
-- profil(2). Also see the VADS documentation on a.prof.



------------------ TRACE CONFIGURATION DATA ----------------------------
    -- The number of records in the event buffer.
    --
    Trace_Buffer_Size : constant Integer := 4096;
    pragma External_Name (Trace_Buffer_Size, "__TRACE_BUFFER_SIZE");

    -- The virtual address of an I/O buffer area if you require a fixed
    -- location.
    -- If this value is null, a buffer will be allocated for you.
    --
    Trace_Buffer_Address : constant System.Address := System.No_Addr;
    pragma External_Name (Trace_Buffer_Address, "__TRACE_BUFFER_ADDRESS");

    ----------------------------------------------------------------------
    -- These symbols are read from your load module by a.db and a.trace
    --
    -- Timer information is now deduced directly from the real_time package.

    ----------------------------------------------------------------------
    -- Kernel events are used internally by run-time developers unlike runtime
    -- events which are used by a.trace analysis. Events and arguments are
    -- subject to change.
    -- If you wish to generate  them, change this boolean to TRUE.
    Kernel_Logging : Boolean := False;
    pragma External_Name (Kernel_Logging, "__KERNEL_LOGGING");
    ----------------------------------------------------------------------





------------------ TASKDEB CONFIGURATION STRUCTURE ---------------------
    -- no_taskdeb: To remove task level debugging from the configuration
    -- remove this section and specify NO_ADDR as the address of a
    -- taskdeb configuration structure below in the configuration table below
    Taskdeb_Config_Rec : Taskdeb_I.Taskdeb_Config_Rec_T :=
       (
        --  INITIALIZATION PROCEDURE ADDRESS:
        Initialize_Proc_Addr => Taskdeb_I.Taskdeb_Initialize'Address,

        --  SIGNAL NUMBER CONFIGURATION:
        Self_Signal => Psignal.Sigusr2,

        --  TRAP INSTRUCTION CONFIGURATION:
        Trap_Instruction_Use => False,    -- not used for self
        Trap_Instruction => 0,
        Trap_Instruction_Length => 0,
        Trap_Intr_Vector_Id => Ada_Krn_Defs.Intr_Vector_Id_T'First,

        --  DEBUG and CALL TASK CONFIGURATION:
        Debug_Task_Stack_Size => 8_000,
        Debug_Task_Attr => No_Addr, -- default task attributes
        Debug_Task_Prio => Priority'Last,  -- highest priority

        Call_Task_Stack_Size => 8_000,
        Call_Task_Attr => No_Addr,        -- default task attributes
        Call_Task_Prio => (Priority'Last - 1));


------------------- CONFIGURATION_TABLE structure -------------------------


-- If you don't want to use the default mutex or condition variable
-- attributes, then, change the V_INIT_ATTR routine to initialize
-- these two attribute records and change the mutex_attr
-- or cond_attr parameter values in the configuration_table
-- from null to mutex_attr'access or cond_attr'access.
    Mutex_Attr : aliased Ada_Krn_Defs.Mutex_Attr_T;
    Cond_Attr : aliased Ada_Krn_Defs.Cond_Attr_T;
    Krn_Task_Attr : aliased Pthread.Pthread_Attr_T;

-- The configuration table can be a constant if the Apex "pragma main(...)"
-- is not required.  All occurances of initial_configuration_table should
-- then be removed or changed to configuration_table.
-- See the procedure V_PRAGMA_MAIN_ADJUST.
    Configuration_Table : V_Usr_Conf_I.
                             Configuration_Table_T; -- remove if constant

-- configuration_table : constant v_usr_conf_i.configuration_table_t
    Initial_Configuration_Table : constant V_Usr_Conf_I.Configuration_Table_T :=
       (

        -- Parameter                  Value
------------                  ---------
        --  STACK CONFIGURATION PARAMETERS:
        Main_Task_Stack_Size => 16#80_0000#,   -- 8MB default for Apex
        Default_Task_Stack_Size => 10_240,
        Exception_Stack_Size => 5000,
        Idle_Stack_Size => 10_240,
        Signal_Task_Stack_Size => 10_240,
        Fast_Rendezvous_Enabled => True,
        Wait_Stack_Size => 4000,
        Zero_Stacks_Enabled => False,

        --  FLOATING-POINT COPROCESSOR CONFIGURATION PARAMETERS:
        Floating_Point_Control =>
           (Rd => V_I_Types.To_Nearest,
            Rp => V_I_Types.Extended,
            Tem => (Invalid | Overflow | Zero_Div => True, others => False),
            Au => False,
            Rs1 => 0,
            Ftt => V_I_Types.None,
            Qne => False,
            Rs2 => 0,
            Fcc => V_I_Types.Equal,
            Aexc => (others => False),
            Cexc => (others => False)),

        C_Floating_Point_Control => (Rd => V_I_Types.To_Nearest,
                                     Rp => V_I_Types.Extended,
                                     Tem => (Invalid => True, others => False),
                                     Au => False,
                                     Rs1 => 0,
                                     Ftt => V_I_Types.None,
                                     Qne => False,
                                     Rs2 => 0,
                                     Fcc => V_I_Types.Equal,
                                     Aexc => (others => False),
                                     Cexc => (others => False)),

        --  HEAP MEMORY CALLOUT CONFIGURATION PARAMETERS:

        Heap_Max => Integer'Last,
        Heap_Extend => 128 * 1024,

        --  MEMORY ALLOCATION CONFIGURATION TABLE:
        Mem_Alloc_Conf_Table_Address => Mem_Alloc_Conf_Table'Address,

        --  TASKDEB CONFIGURATION structure:
        --     no_taskdeb: To remove task level debugging from the configuration
        --     assign TASKDEB_CONFIG_ADDRESS to NO_ADDR and copletely remove
        --     the taskdeb configuration record defined above.
        Taskdeb_Config_Address => Taskdeb_Config_Rec'Address,
        -- UNIX SIGNAL CONFIGURATION PARAMETERS:
        Disable_Signals_Mask => (16#7E3FB007#, 16#FFE#, 16#0#, 16#0#),
        Numeric_Signal_Enabled => True,
        Storage_Signal_Enabled => False,
        --  Solaris MT Ada CONFIGURATION PARAMETERS:
        Concurrency_Level => 1,
        Enable_Signals_Mask => (16#7E3FB007#, 16#FFE#, 16#0#, 16#0#),
        Exit_Signals_Mask => (16#18006#, others => 16#0#),
        Intr_Task_Prio => Any_Priority'Last,
        Intr_Task_Stack_Size => 10_240,

        --  ATTRIBUTES CONFIGURATION PARAMETERS:
        Default_Task_Attributes => (Prio => Priority'First,
                                    Krn_Task_Attr => null,

                                    Mutex_Attr => null,   -- use default mutex attr
                                    Cond_Attr => null),      -- use default cond attr

        Main_Task_Attr_Address => No_Addr,    -- use DEFAULT_TASK_ATTRIBUTES
        Signal_Task_Attr_Address => No_Addr,    -- use DEFAULT_TASK_ATTRIBUTES
        Intr_Task_Attr_Address => No_Addr,    -- use DEFAULT_TASK_ATTRIBUTES
        Masters_Mutex_Attr_Address => No_Addr,    -- use default mutex attr
--                         =>      mutex_attr'address,
        Mem_Alloc_Mutex_Attr_Address => No_Addr,    -- use default mutex attr
--                         =>      mutex_attr'address,
        Ada_Io_Mutex_Attr_Address => No_Addr,    -- use default mutex attr
--                         =>      mutex_attr'address,

        --  CODE COVERAGE CONFIGURATION PARAMETERS:
        Send_Sigint_To_Cca => True,

        --  MISC CONFIGURATION PARAMETERS:
        Old_Style_Max_Intr_Entry => System.Address'Ref (16#1FF#), -- 512 - 1
        Task_Storage_Size => 64

        );


    pragma External_Name (Initial_Configuration_Table,
                          "USR_CONFIGURATION_TABLE");
    -- If pragma main is not required then "configuration_table" should
    -- be constant and given the external name USR_CONFIGURATION_TABLE like:
    --   pragma external_name(configuration_table, "USR_CONFIGURATION_TABLE");
    -- Note: above external_name is needed by a.ld for building
    -- the link_block.  Currently two of the statically initialized fields
    -- are used during kernel initialization:
    --   MAIN_TASK_STACK_SIZE
    --   EXCEPTION_STACK_SIZE




    Stop : System.Address;
    pragma Import (Ada, Stop, Link_Name => "_STOP");
-- The top of stack sentinel, accessed through v_usr_conf_i.  See
-- the EXCEPTION_UNWIND_SAFE package.

-- Variable where the OS RLIMIT_DATA process data segment maximum is stored.
    type Rlim_T is new System.Storage_Elements.Integer_Address;
    Data_Limit : Rlim_T := Rlim_T'Last;



    V_Start_Program_Common_Called : Boolean := False;
    V_Adainit_Called : Boolean := False;
    V_Adafinal_Called : Boolean := False;
    Adafinal_Callout_Address : Address;



--------------------- EXCEPTION_UNWIND_SAFE package --------------------
    package Exception_Unwind_Safe is
        -- Package used to call initialization procedures and tables
        -- of procedures (.ie elaboration table).  Unhandled exceptions
        -- are handled correctly for the procedures called.

        -- Function's below return an unconstrained array.
        type Array_Int is array (Positive range <>) of Integer;

        -- Initialize this package.
        procedure Init;

        -- Call a parameterless procedure.  Fail correctly if the
        -- procedure raises an unhandled exception.
        function Call_Proc (Proc : Address) return Integer;

        -- Call a table of elaboration procedures.  The address of a null
        -- terminated table is passed as a parameter.  Return an unconstrained
        -- array which has a single element that contains the return value
        -- from the last procedure called.
        --
        -- The label __start_continue is a sentinal value for the debugger.
        -- It triggers the debugger where to stop unwinding.
        --
        -- call_proc_table_minus_one does not call the last non-null
        -- entry in the table.
        --
        function Call_Proc_Table (Table : Address) return Array_Int;
        pragma External_Name (Call_Proc_Table, "__start_continue");
        function Call_Proc_Table_Minus_One (Table : Address) return Array_Int;
        -- For ada95
        function Call_Proc_Table_Minus_Number
                    (Table : Address; Num : Integer) return Array_Int;
        function Call_Proc_Table_Rest
                    (Table : Address; Num : Integer) return Array_Int;
    end Exception_Unwind_Safe;

    package body Exception_Unwind_Safe is
        subtype Array_Int_1 is Array_Int (1 .. 1);

        -- Move the return address to the stop label v_usr_conf_i.stop.
        -- This label is used as a sentinal value for the exception unwinding
        -- code which will unwind until the return pc value is the value
        -- in v_usr_conf_i.stop, at which point the program (or the debugger)
        -- will announce "unhandled exception ..."
        --
        procedure Move_Return_Address_To_Stop is
        begin
            V_Usr_Conf_I.Stop := System.Return_Address;
        end Move_Return_Address_To_Stop;

        -- The initialization routine for the package.
        -- All intialization procedures that need to announce unhandled
        -- exceptions must call through call_proc_table.
        -- This initialization routine sets the value of v_usr_conf_i.stop
        procedure Init is
            Table : array (1 .. 2) of Address;
            Ret : Integer;
        begin
            Table (1) := Move_Return_Address_To_Stop'Address;
            Table (2) := No_Addr;
            Ret := Call_Proc_Table (Table'Address) (1);
        end Init;

        -- Call a single procedure, announce unhandled exceptions.
        function Call_Proc (Proc : Address) return Integer is
            Table : array (1 .. 2) of Address;
        begin
            Table (1) := Proc;
            Table (2) := No_Addr;
            return Call_Proc_Table (Table'Address) (1);
        end Call_Proc;

        -- Call all the procedures in a null terminated array of procedure
        -- addresses, announce unhandled exceptions.  All initialization
        -- procedures must be called throgh this function.
        --
        -- This function must return an unconstrained array.  This will cause
        -- the compiler to use the frame pointer in this procedure.  This is
        -- required since the procedures being called (in the case of some
        -- generics) will grow the stack (.ie stack creep).
        --
        -- Returning an unconstrained array will also cause the caller of
        -- this procedure to recognize that the stack pointer will not
        -- be restored upon return from this function
        --
        function Call_Proc_Table (Table : Address) return Array_Int is
            type A_Address is access Address;
            for A_Address'Storage_Size use 0;
            function To_A_Address is
               new Unchecked_Conversion (Address, A_Address);
            function To_Address is
               new Unchecked_Conversion (A_Address, Address);
            Proc : A_Address := To_A_Address (Table);
            Ret : Integer := 0;
        begin
            while Proc.all /= No_Addr loop
                declare
                    function Integer_Function_Integer_Param
                                (P : Integer) return Integer;
                    pragma Interface (Ada, Integer_Function_Integer_Param);
                    for Integer_Function_Integer_Param'Address use Proc.all;
                begin
                    Ret := Integer_Function_Integer_Param (Ret);
                end;
                Proc := To_A_Address (To_Address (Proc) +
                                      Integer'(Proc.all'Size /
                                               System.Storage_Unit));
            end loop;
            return Array_Int_1'(1 => Ret);
        end Call_Proc_Table;

        -- See documentation above for call_proc_table.
        -- This routine must call call_proc_table.
        --
        function Call_Proc_Table_Minus_One (Table : Address) return Array_Int is
            Proc_Table : V_Usr_Data.Elaboration_Table_T;
            for Proc_Table'Address use Table;
            Table_Size : Integer := 1;
        begin
            while Proc_Table (Table_Size) /= No_Addr loop
                Table_Size := Table_Size + 1;
            end loop;
            Table_Size := Table_Size - 1;    -- make this table one smaller
            if Table_Size <= 1 then
                return Array_Int_1'(1 => 0);
            end if;
            declare
                type Temp_Elaboration_Table_T is
                   array (1 .. Table_Size) of Address;
                Temp_Table : Temp_Elaboration_Table_T;
            begin
                Temp_Table (1 .. Table_Size - 1) :=
                   Temp_Elaboration_Table_T (Proc_Table (1 .. Table_Size - 1));
                Temp_Table (Table_Size) := No_Addr;
                return Call_Proc_Table (Temp_Table'Address);
            end;
        end Call_Proc_Table_Minus_One;

        -- For ada95
        function Call_Proc_Table_Minus_Number
                    (Table : Address; Num : Integer) return Array_Int is
            Proc_Table : V_Usr_Data.Elaboration_Table_T;
            for Proc_Table'Address use Table;
            Table_Size : Integer := 1;
        begin
            while Proc_Table (Table_Size) /= No_Addr loop
                Table_Size := Table_Size + 1;
            end loop;
            Table_Size := Table_Size - Num; -- make this table one smaller
            if Table_Size <= 1 then
                return Array_Int_1'(1 => 0);
            end if;
            declare
                type Temp_Elaboration_Table_T is
                   array (1 .. Table_Size) of Address;
                Temp_Table : Temp_Elaboration_Table_T;
            begin
                Temp_Table (1 .. Table_Size - 1) :=
                   Temp_Elaboration_Table_T (Proc_Table (1 .. Table_Size - 1));
                Temp_Table (Table_Size) := No_Addr;
                return Call_Proc_Table (Temp_Table'Address);
            end;
        end Call_Proc_Table_Minus_Number;

        -- See documentation above for call_proc_table.
        -- This routine must call call_proc_table.
        -- For ada95
        --
        function Call_Proc_Table_Rest
                    (Table : Address; Num : Integer) return Array_Int is
            Proc_Table : V_Usr_Data.Elaboration_Table_T;
            for Proc_Table'Address use Table;
            Table_Size : Integer := 1;
            Full_Table_Size : Integer;
        begin
            while Proc_Table (Table_Size) /= No_Addr loop
                Table_Size := Table_Size + 1;
            end loop;

            Full_Table_Size := Table_Size - 1;
            Table_Size := Table_Size - Num; -- make this table one smaller
            if Table_Size <= 1 or Num = 0 then
                return Array_Int_1'(1 => 0);
            end if;
            declare
                type Temp_Elaboration_Table_T is
                   array (1 .. Num + 1) of Address;
                Temp_Table : Temp_Elaboration_Table_T;
            begin
                Temp_Table (1 .. Num) :=
                   Temp_Elaboration_Table_T
                      (Proc_Table (Table_Size .. Full_Table_Size));
                Temp_Table (Num + 1) := No_Addr;
                return Call_Proc_Table (Temp_Table'Address);
            end;
        end Call_Proc_Table_Rest;

    end Exception_Unwind_Safe;


--------------------- V_SIGNAL_ISR routine  -------------------------------

    procedure V_Signal_Isr (I : V_I_Sig.Isr_Header_Ref) is
        -- The compiler emits code to call this routine for an interrupt entry
        -- in an active (non-passive) task. The interrupt entry acts as a signal
        -- that is posted when the interrupt occurs.
    begin
        -- Upon entry:
        --
        --  Scratch registers saved, STACK_LIMIT switched and top of exception
        --  unwinding set to top of ISR (versus top of program).

        -- Post interrupt entry's signal
        V_I_Sig.Post_Signal (I.S);
    end V_Signal_Isr;


--------------------- V_CIFO_ISR routine  -----------------------------

    procedure V_Cifo_Isr (I : V_I_Cifo.Cifo_Isr_Header_Ref) is

        -- The compiler emits code to call this routine for an interrupt entry
        -- in a task using the CIFO INTERRUPT_TASK pragma. This routine
        -- is called with a pointer to a record containing the interrupt
        -- vector number and the address of the interrupt entry's
        -- accept body. Check the CIFO documentation for the restrictions
        -- imposed on an interrupt task.
        --
        -- Note: this routine is called for all CIFO interrupt task ISR's.
        --

        use Machine_Code;
    begin
        -- Upon entry:
        --
        --  Scratch registers saved, STACK_LIMIT switched and top of exception
        --  unwinding set to top of ISR (versus top of program).

        -- Call the accept body that has been compiled like a normal
        -- procedure.
        Code_2'(Ld, Addr (I0 + V_I_Cifo.Cifo_Isr_Handler_Off), O0);
        Code_2'(Jmpl, O0, O7);
        Code_0'(Op => Nop);
    end V_Cifo_Isr;


------------------- STORAGE_MANAGEMENT_CALLOUT routines ------------------

    -- The default external storage pool implementation will use the
    -- three callouts given below.  If MALLOC_BASED_ALLOCATION has
    -- been configured then all allocation/deallocation requests are
    -- routed through these callouts - where the default callout
    -- implementation is malloc/free.  If MALLOC_BASED_ALLOCATION has
    -- not been configured then the allocate callout is only used
    -- when the default storage pool needs to be extended.  In this
    -- case the malloc/free implementation could be replaced by a
    -- sbrk/null alternative.



    function V_Extern_Alloc_Callout (Size : Alloc_T) return Address is

        function Malloc (Size : Alloc_T) return Address;
        pragma Import (C, Malloc, "malloc");

        procedure Free (A : System.Address);
        pragma Import (C, Free, "free");

        -- Although normally not required, some OS environments require
        -- some amount of malloc space to always be available.
        Reserve_Space : constant := 16#0#;
        Base : System.Address;
    begin
        if Reserve_Space /= 0 then
            Base := Malloc (Size + Reserve_Space);
            if Base /= No_Addr then
                Free (Base);
                Base := Malloc (Size);
            end if;
        else
            Base := Malloc (Size);
        end if;
        return Base;
    end V_Extern_Alloc_Callout;

    procedure V_Extern_Free_Callout (A : Address) is
        procedure Free (Mem : Address);
        pragma Import (C, Free, "free");
    begin
        Free (A);
    end V_Extern_Free_Callout;


    function V_Extern_Size_Callout return Alloc_T is
    begin
        if Data_Limit > Rlim_T (Alloc_T'Last) then
            return Alloc_T'Last;
        else
            return Alloc_T (Data_Limit);
        end if;
    end V_Extern_Size_Callout;


    -- Memory allocations/deallocations in the kernel are supported by
    -- the callouts below.  The default implementation makes calls to
    -- the default kernel_external_storage_pool (configured in the
    -- STORAGE_POOL_CONFIGURATION_TABLE).  The implementations of
    -- these callouts will need to be modified if: (1) a non-default
    -- kernel_external_storage_pool is configured that does not
    -- provide kernel_allocate, kernel_deallocate, and kernel_pool_init
    -- routines (with the external names given below); or (2) the 
    -- application wishes to use a different storage management scheme 
    -- for the kernel, e.g., use malloc/free directly.

    function V_Krn_Alloc_Callout (Size : Integer) return Address is
        function Kernel_Allocate (Size : Alloc_T) return System.Address;
        pragma Import (Ada, Kernel_Allocate,
                       Link_Name => "__USR_CONF_KRN_ALLOCATE");
        Base : Address;
    begin
        Base := Kernel_Allocate (Size);
        if Base = No_Addr then
            Base := V_I_Types.Integer_To_Address (-1);
        end if;
        return Base;
    end V_Krn_Alloc_Callout;

    procedure V_Krn_Free_Callout (A : Address) is
        procedure Kernel_Deallocate (A : System.Address);
        pragma Import (Ada, Kernel_Deallocate,
                       Link_Name => "__USR_CONF_KRN_DEALLOCATE");
    begin
        Kernel_Deallocate (A);
    end V_Krn_Free_Callout;

    procedure V_Krn_Pool_Init_Callout is
        procedure Kernel_Pool_Init;
        pragma Import (Ada, Kernel_Pool_Init,
                       Link_Name => "__USR_CONF_KRN_POOL_INITIALIZE");
    begin
        Kernel_Pool_Init;
    end V_Krn_Pool_Init_Callout;

    procedure V_Krn_Pool_Finalize_Callout is
        procedure Kernel_Pool_Finalize;
        pragma Import (Ada, Kernel_Pool_Finalize,
                       Link_Name => "__USR_CONF_KRN_POOL_FINALIZE");
    begin
        Kernel_Pool_Finalize;
    end V_Krn_Pool_Finalize_Callout;

---------- V_START_PROGRAM and V_START_PROGRAM_CONTINUE routines ---------


    procedure V_Init_Attr (Configuration_Table :
                              V_Usr_Conf_I.A_Configuration_Table_T) is

        use type Interfaces.C.Int;
        use type Pthread.A_Pthread_Attr_T;
        use type V_I_Main.Task_Dispatching_Policy_Type;
        use type V_I_Main.Locking_Policy_Type;


        -- PATCH
        -- Add Pthread_Attr_Setscope, missing from Apex 3.2.0b package Pthread.
        function Pthread_Attr_Setscope (Attr : Pthread.A_Pthread_Attr_T;
                                        Contentionscope : Interfaces.C.Int)
                                       return Interfaces.C.Int;
        pragma Import (C, Pthread_Attr_Setscope, "pthread_attr_setscope");

        Pthread_Scope_Process : constant := 0;
        Pthread_Scope_System : constant := 1;

        function To_String (Cptr : Unix_Types.Char_Ptr) return String is  
            use type Unix_Types.Char_Ptr;

            function To_Character is new Unchecked_Conversion (Unix_Types.Char,  
                                                               Character);
            pragma Suppress (Access_Check);
            pragma Suppress (Range_Check);
            pragma Suppress (Index_Check);
        begin
            if Cptr /= null then
                declare
                    Rover : Unix_Types.Char_Ptr := Cptr;
                    C : Natural := Unix_Types.Length (Cptr);
                    S : String (1 .. C);
                begin
                    for I in 1 .. C loop
                        S (I) := To_Character (Rover.all);
                        Rover := Unix_Types.Increment (Rover);
                    end loop;
                    return S;
                end;
            end if;
            return "";
        end To_String;

        function Unix_Get_Error_String
                    (Status : Interfaces.C.Int) return String is
        begin
            return To_String (Unix_Operations.Strerror
                                 (Unix_Types.Int (Status)));
        end Unix_Get_Error_String;

        New_Task : Task_Id;
        New_Thread : aliased Pthread.Pthread_T;
        Os_Status : Interfaces.C.Int;
        Attr : aliased Pthread.Pthread_Attr_T;
        A_Attr : Pthread.A_Pthread_Attr_T;
        Policy : Interfaces.C.Int;
        Param : aliased Pthread.Sched_Param;


        subtype Extended_Prio is Integer range System.Priority'First - 1 ..
                                                  System.Any_Priority'Last + 1;
        Unspecified_Prio : Extended_Prio := System.Any_Priority'Last + 1;

        function To_Priority is new Unchecked_Conversion
                                       (Extended_Prio, System.Priority);

        -- If you don't want to use the default mutex or condition variable
        -- attributes, then, change this routine to initialize the
        -- mutex_attr or cond_attr records and change the mutex_attr_address
        -- or cond_attr_address parameter values in the configuration_table
        -- from NO_ADDR to mutex_attr'address or cond_attr'address.


    begin
        --
        -- You can use one of the attr_init routines in ada_krn_defs as follows:
        --
        --    ada_krn_defs.fifo_mutex_attr_init(attr => mutex_attr'access);
        --                or
        --    ada_krn_defs.prio_mutex_attr_init(attr => mutex_attr'access);
        --                or
        --    ada_krn_defs.prio_inherit_mutex_attr_init(attr => mutex_attr'access);
        --                or
        --    ada_krn_defs.prio_ceiling_mutex_attr_init(attr => mutex_attr'access,
        --        ceiling_prio=> 80);

        -- Set the Prio component to a sentinel to tell the runtime system
        -- to set the priority according to the value specified by the
        -- compiler (from the Priority pragam or by default).

        Configuration_Table.Default_Task_Attributes.Prio :=
           To_Priority (Unspecified_Prio);

        A_Attr := Krn_Task_Attr'Access;
        Os_Status := Pthread.Pthread_Attr_Init (A_Attr);
        if Os_Status /= 0 then
            Ada_Krn_I.Panic_Exit ("pthread_attr_init failed, OS says: " &
                                  Unix_Get_Error_String (Os_Status));
        end if;
        Configuration_Table.Default_Task_Attributes.Krn_Task_Attr := A_Attr;


        -- Uncomment this to cause all tasks (threads) to be scheduled against
        -- all other threads in the system.
        -- Os_Status := Pthread_Attr_Setscope (A_Attr, Pthread_Scope_System);
        -- if Os_Status /= 0 then
        --     Ada_Krn_I.Panic_Exit (
        --         "pthread_attr_setscope failed, OS says: " &
        --         Unix_Get_Error_String (Os_Status));
        -- end if;


        null;
    end V_Init_Attr;

    -- Remove this procedure if not supporting apex pragma main.
    -- Adjust the final configuration table with the values from pragma main.
    procedure V_Pragma_Main_Adjust
                 (Initial_Configuration_Table :
                     V_Usr_Conf_I.A_Configuration_Table_T;
                  Configuration_Table : V_Usr_Conf_I.A_Configuration_Table_T) is
    begin
        Configuration_Table.all := Initial_Configuration_Table.all;
        if V_I_Main.Stack_Size /= -1 then
            Configuration_Table.Main_Task_Stack_Size := V_I_Main.Stack_Size;
        end if;

        if V_I_Main.Exception_Stack_Size /= -1 then  
            Configuration_Table.Exception_Stack_Size :=
               V_I_Main.Exception_Stack_Size;
        end if;

        if V_I_Main.Default_Task_Stack_Size /= -1 then
            Configuration_Table.Default_Task_Stack_Size :=
               V_I_Main.Default_Task_Stack_Size;
        end if;

        if V_I_Main.Default_Task_Priority >= 0 then
            Configuration_Table.Default_Task_Attributes.Prio :=
               V_I_Main.Default_Task_Priority;
        end if;

        if V_I_Main.Fuzzy."=" (V_I_Main.Fast_Rendezvous_Enabled,
                               V_I_Main.Fuzzy.True) then
            Configuration_Table.Fast_Rendezvous_Enabled := True;
        elsif V_I_Main.Fuzzy."=" (V_I_Main.Fast_Rendezvous_Enabled,
                                  V_I_Main.Fuzzy.False) then
            Configuration_Table.Fast_Rendezvous_Enabled := False;
        end if;

        if V_I_Main.Idle_Stack_Size /= -1 then  
            Configuration_Table.Idle_Stack_Size := V_I_Main.Idle_Stack_Size;
        end if;

        if V_I_Main.Signal_Task_Stack_Size /= -1 then  
            Configuration_Table.Signal_Task_Stack_Size :=
               V_I_Main.Signal_Task_Stack_Size;
        end if;

        if V_I_Main.Wait_Stack_Size /= -1 then  
            Configuration_Table.Wait_Stack_Size := V_I_Main.Wait_Stack_Size;
        end if;


        if V_I_Main.Heap_Extend /= -1 then  
            Configuration_Table.Heap_Extend := V_I_Main.Heap_Extend;
        end if;


        if V_I_Main.Concurrency_Level /= -1 then  
            Configuration_Table.Concurrency_Level := V_I_Main.Concurrency_Level;
        end if;

        if V_I_Main.Intr_Task_Prio /= -1 then
            Configuration_Table.Intr_Task_Prio := V_I_Main.Intr_Task_Prio;
        end if;
        if V_I_Main.Intr_Task_Stack_Size /= -1 then  
            Configuration_Table.Intr_Task_Stack_Size :=
               V_I_Main.Intr_Task_Stack_Size;
        end if;

        if V_I_Main.Fuzzy."=" (V_I_Main.Numeric_Signal_Enabled,
                               V_I_Main.Fuzzy.True) then  
            Configuration_Table.Numeric_Signal_Enabled := True;
        elsif V_I_Main.Fuzzy."=" (V_I_Main.Numeric_Signal_Enabled,
                                  V_I_Main.Fuzzy.False) then
            Configuration_Table.Numeric_Signal_Enabled := False;
        end if;

        if V_I_Main.Fuzzy."=" (V_I_Main.Storage_Signal_Enabled,
                               V_I_Main.Fuzzy.True) then
            Configuration_Table.Storage_Signal_Enabled := True;
        elsif V_I_Main.Fuzzy."=" (V_I_Main.Storage_Signal_Enabled,
                                  V_I_Main.Fuzzy.False) then
            Configuration_Table.Storage_Signal_Enabled := False;
        end if;

        if V_I_Main.Task_Storage_Size /= -1 then  
            Configuration_Table.Task_Storage_Size := V_I_Main.Task_Storage_Size;
        end if;


        if V_I_Main.Fuzzy."=" (V_I_Main.Zero_Stacks_Enabled,
                               V_I_Main.Fuzzy.True) then
            Configuration_Table.Zero_Stacks_Enabled := True;
        elsif V_I_Main.Fuzzy."=" (V_I_Main.Zero_Stacks_Enabled,
                                  V_I_Main.Fuzzy.False) then
            Configuration_Table.Zero_Stacks_Enabled := False;
        end if;

    end V_Pragma_Main_Adjust;

    procedure Do_Nothing (Addr : System.Address) is
    begin
        -- This procedure inhibits the reference to symbol_list in
        -- force_symbols_into_transitive_closure from being optimized away.
        null;
    end Do_Nothing;
    pragma Inline_Never (Do_Nothing);


    procedure Force_Symbols_Into_Transitive_Closure is
        Ada_Main_Slot : Integer;
        pragma Interface_Name (Ada_Main_Slot, "__ADA_MAIN_SLOT");



        Adapath_External : Address;
        pragma Import (Ada, Adapath_External,
                       Link_Name => "__ADAPATH_external");



        type Symbol_List_Array is array (Positive range <>) of System.Address;
        Symbol_List : constant Symbol_List_Array :=
           (Ada_Main_Slot'Address,
            -- adapath_external is used internally at rational
            -- to generate cannonical view names from a linked
            -- executable.  This is optional at customer sites.
            Adapath_External'Address, System.No_Addr);

    begin
        -- This procedure simply forces a reference to any symbols that
        -- are needed into the executable which aren't in the transitive
        -- closure of __start.  Currently these symbols are used only by the
        -- debugger, but others may be added in the future for other purposes.
        Do_Nothing (Symbol_List'Address);
    end Force_Symbols_Into_Transitive_Closure;


    ---
    procedure V_Init_Usr_Data (Configuration_Table :
                                  V_Usr_Conf_I.A_Configuration_Table_T) is
        use V_I_Main;

        Local_Main_Global_Entry_Criteria : V_I_Cifo.Queuing_T;
    begin

        Force_Symbols_Into_Transitive_Closure;

        -- Use the global_entry_criteria defined by
        -- "pragma Main (Queuing_Policy => ...)".
        -- This is supposed to represent the Ada 95 Queuing_Policy pragma,
        -- which is a configuration pragma, effecting the entire partition.
        -- For Ada 95, the ability to have different queueing policies in the
        -- same partition is not used (at least not yet).

        case V_I_Main.Queuing_Policy is
            when V_I_Main.Unspecified =>
                Local_Main_Global_Entry_Criteria :=
                   V_I_Cifo.Main_Global_Entry_Criteria;
            when V_I_Main.Fifo_Queuing =>
                Local_Main_Global_Entry_Criteria := V_I_Cifo.Fifo_Queuing;
            when V_I_Main.Priority_Queuing =>
                Local_Main_Global_Entry_Criteria := V_I_Cifo.Priority_Queuing;
        end case;


        -- Our current implementation of fast rendezvous doesn't support
        -- Fifo_Within_Priorities; it handles loss of inherited priority by
        -- the server task incorrectly, putting it at the tail of the queue for
        -- the new priority rather than the head.  If the user hasn't explicitly
        -- asked for fast rendezvous using pragma Main, turn it off.

        if V_I_Main.Task_Dispatching_Policy =
           V_I_Main.Fifo_Within_Priorities then
            if V_I_Main.Fuzzy."=" (V_I_Main.Fast_Rendezvous_Enabled,
                                   V_I_Main.Fuzzy.Unspecified) then
                Configuration_Table.Fast_Rendezvous_Enabled := False;
            end if;
        end if;

        -- Our current implementation of fast rendezvous doesn't support
        -- Fifo_Within_Priorities; it handles loss of inherited priority by
        -- the server task incorrectly, putting it at the tail of the queue for
        -- the new priority rather than the head.  If the user hasn't explicitly
        -- asked for fast rendezvous using pragma Main, turn it off.

        if V_I_Main.Task_Dispatching_Policy =
           V_I_Main.Fifo_Within_Priorities then
            if V_I_Main.Fuzzy."=" (V_I_Main.Fast_Rendezvous_Enabled,
                                   V_I_Main.Fuzzy.Unspecified) then
                Configuration_Table.Fast_Rendezvous_Enabled := False;
            end if;
        end if;

        V_Usr_Data.Main_Pragmas :=
           (Main_Task_Priority => V_I_Cifo.Main_Task_Priority,
            Priority_Inheritance_Enabled =>
               V_I_Cifo.Main_Priority_Inheritance_Enabled = 1,
            Global_Entry_Criteria => Local_Main_Global_Entry_Criteria,
            Global_Select_Lexical_Order =>
               V_I_Cifo.Main_Global_Select_Lexical_Order = 1);


        V_Usr_Data.Default_Task_Stack_Size :=
           Configuration_Table.Default_Task_Stack_Size;

        V_Usr_Data.Fp_Control_For_Ada :=
           Configuration_Table.Floating_Point_Control;
        V_Usr_Data.Fp_Control_For_C :=
           Configuration_Table.C_Floating_Point_Control;

    end V_Init_Usr_Data;


    -- This version of UNIX panic exit is for use before any kernel
    -- initialization has taken place.
    procedure Uninitialized_Panic_Exit (S : String; Status : Integer := -1) is
        procedure Unix_Exit (Status : Integer);
        pragma Import (C, Unix_Exit, "exit");
        procedure Write (Fd : Integer; Buf : Address; Len : Integer);
        pragma Interface (C, Write);
    begin
        Write (1, S (S'First)'Address, S'Length);
        Unix_Exit (Status);
    end Uninitialized_Panic_Exit;


    procedure Set_Resource_Limits is

        use System.Storage_Elements;


        Rlimit_Data : constant Integer := 2;
        Rlimit_Stack : constant Integer := 3;
        -- setting the main task stack limit on the self can be a little loose
        -- (self systems have memory to spare).
        Main_Stack_Padding : constant Integer :=
           16#1_000# +  -- stack limit is derived from the current not absolute
              16#2_000#;   -- round to next 8k boundry


        type Rlimit_T is
            record
                Cur_Limit : Rlim_T;
                Max_Limit : Rlim_T;
            end record;
        type A_Rlimit_T is access all Rlimit_T;
        for A_Rlimit_T'Storage_Size use 0;

        Rlimit : aliased Rlimit_T;

        Padded_Stack_Limit : constant Rlim_T :=
           Rlim_T (Configuration_Table.Main_Task_Stack_Size +
                   Configuration_Table.Exception_Stack_Size +
                   Main_Stack_Padding);

        Status : Integer;

        function Get_Error_String return String is

            function Strerror (Errnum : Integer) return System.Address;
            pragma Import (C, Strerror);
            The_Char_Ptr : System.Address;

        begin
            The_Char_Ptr := Strerror (System.Errno_Support.Get_Errno);
            return V_I_Libop.To_String (The_Char_Ptr);
        end Get_Error_String;

        function Getrlimit
                    (Resource : Integer; Rlimit : A_Rlimit_T) return Integer;
        pragma Interface (C, Getrlimit);

        function Setrlimit
                    (Resource : Integer; Rlimit : A_Rlimit_T) return Integer;
        pragma Interface (C, Setrlimit);

    begin
        Status := Getrlimit (Rlimit_Data, Rlimit'Access);

        if Configuration_Table.Heap_Max /= Integer'Last then
            -- set the data size to what was in the configuration table.
            -- pragma main may have altered this value.
            Rlimit.Cur_Limit := Rlim_T (Configuration_Table.Heap_Max);
            Status := Setrlimit (Rlimit_Data, Rlimit'Access);
            if Status /= 0 then
                Uninitialized_Panic_Exit
                   ("Error: encountered problem setting DATA limit to " &
                    "V_USR_CONF.CONFIGURATION_TABLE.HEAP_MAX's value of " &
                    Integer'Image (Integer (Configuration_Table.Heap_Max)) &
                    Ascii.Cr & Ascii.Lf & "OS says: " &
                    Get_Error_String & Ascii.Cr & Ascii.Lf & Ascii.Nul);
            end if;
        end if;
        Data_Limit := Rlimit.Cur_Limit;

        Status := Getrlimit (Rlimit_Stack, Rlimit'Access);

        -- padded_stack_limit should only be less than zero if it
        -- has wrapped around, which means that more stack has been
        -- requested than can fit in an integer.
        if (Padded_Stack_Limit > Rlimit.Max_Limit) or else
           (Padded_Stack_Limit < 0) then
            Uninitialized_Panic_Exit ("Error: Requested stack size of" &
                                      Rlim_T'Image (Padded_Stack_Limit) &
                                      " exceeds system limit of" &
                                      Rlim_T'Image (Rlimit.Max_Limit) &
                                      Ascii.Cr & Ascii.Lf & Ascii.Nul);
        end if;

        if Padded_Stack_Limit > Rlimit.Cur_Limit then
            -- Increase the size of the stack area
            Rlimit.Cur_Limit := Rlim_T (Padded_Stack_Limit);
            Status := Setrlimit (Rlimit_Stack, Rlimit'Access);
            if Status /= 0 then
                Uninitialized_Panic_Exit
                   ("Error: encountered problem setting STACK limit to " &
                    Rlim_T'Image (Padded_Stack_Limit) &
                    Ascii.Cr & Ascii.Lf & "OS says: " &
                    Get_Error_String & Ascii.Cr & Ascii.Lf & Ascii.Nul);
            end if;
        end if;
    end Set_Resource_Limits;




    -- THE FOLLOWING ROUTINES SHOULD NOT BE MODIFIED

    procedure V_Start_Program_Common (Startup_Continue : Address)
                 -- Common startup for self and cross.
                 -- Do initial program configuration and intialize the runtime.
                 -- Return through the procedure defined in startup_continue 
                 is
        Ret : Integer;
    begin
        if not V_Start_Program_Common_Called then
            V_Start_Program_Common_Called := True;
        else
            return;
        end if;

        -- Adjust the configuration table with pragma main values
        -- Remove this if apex pragma main is not going to be supported
        -- and the configuration_table is a constant.
        V_Pragma_Main_Adjust (Initial_Configuration_Table =>
                                 V_Usr_Conf_I.To_A_Configuration_Table_T
                                    (Initial_Configuration_Table'Address),
                              Configuration_Table =>
                                 V_Usr_Conf_I.To_A_Configuration_Table_T
                                    (Configuration_Table'Address));


        -- Set current data limit so that its less than HEAP_MAX. On
        -- some OSs the program hangs or is killed if sbrk's cause
        -- the data area to get too big.
        --
        -- Also, change the current stack limit if its less than
        -- MAIN_STACK_SIZE + EXCEPTION_STACK_SIZE
        V_Init_Usr_Data (V_Usr_Conf_I.To_A_Configuration_Table_T
                            (Configuration_Table'Address));
        Set_Resource_Limits;
        -- Initialize tasking systems storage management
        V_Krn_Pool_Init_Callout;
        V_Init_Attr (V_Usr_Conf_I.To_A_Configuration_Table_T
                        (Configuration_Table'Address));
        Ret := V_Usr_Conf_I.Ts_Initialize
                  (Usr_Link_Block => Link_Block.To_A_Link_Block_T
                                        (V_Usr_Conf_I.Usr_Link_Block'Address),
                   Configuration_Table => V_Usr_Conf_I.
                                             To_A_Configuration_Table_T
                                             (Configuration_Table'Address),
                   Main_Pragmas => V_I_Cifo.To_A_Main_Pragmas_T
                                      (V_Usr_Data.Main_Pragmas'Address),
                   Startup_Continue => Startup_Continue);
    end V_Start_Program_Common;

    procedure V_Start_Program is
    begin
        V_Start_Program_Common (V_Start_Program_Continue'Address);
    end V_Start_Program;

    procedure V_Start_Program_Self
                 (Argc : Integer; Argv : Address; Envp : Address) is
    begin
        V_Usr_Conf_I.Argc_Value := Argc;
        V_Usr_Conf_I.Argv_Value := Argv;
        -- v_usr_conf_i.environ := envp; done in crt0
        V_Start_Program;

        -- No return.

    end V_Start_Program_Self;

    procedure V_Start_Program_Continue_Ada_1 is
        -- Initialization required as a pre-requisite to VADS C++ initialization.
    begin
        null;
    end V_Start_Program_Continue_Ada_1;

    procedure V_Start_Program_Continue_Ada_2 is
        -- Initialization during ADA program startup or on the thread
        -- of V_ADAINIT
    begin
        Exception_Unwind_Safe.Init;

        declare

            -- Convert ascii into integer. Must be a positive number.
            procedure A_To_I (Text : String;
                              Num : out Integer;
                              Cvt_Ok : out Boolean) is
                Max_Sum : constant Integer := Integer'Last / 10;
                Last_Digit : constant Integer := Integer'Last rem 10;
                Digit : Integer;
                Neg : Boolean := False;
                Sum : Integer := 0;
            begin
                Cvt_Ok := True;
                for I in Text'Range loop
                    Digit := Character'Pos (Text (I)) - Character'Pos ('0');
                    if Digit in 0 .. 9 and then
                       ((Sum < Max_Sum) or else (Sum = Max_Sum and then
                                                 Digit <= Last_Digit)) then
                        Sum := Sum * 10 + Digit;
                    else
                        Cvt_Ok := False;
                        exit;
                    end if;
                end loop;
                Num := Sum;
            end A_To_I;

            -- If the configuration table's CONCURRENCY_LEVEL equals 0 or 1,
            -- allow the environment variable, ADA_CONCURRENCY_LEVEL, to
            -- override it.
            procedure Concurrency_Init is
                Concurrency_Env_Variable_String : constant String (1 .. 22) :=
                   "ADA_CONCURRENCY_LEVEL" & Ascii.Nul;
                Ada_Concurrency_Level_Error : exception;

                type A_String is access String (1 .. 10000);
                for A_String'Storage_Size use 0;
                function Getenv (Ev : Address) return A_String;
                pragma Import (C, Getenv, External_Name => "getenv");

                Concurrency_Level : Natural :=
                   Configuration_Table.Concurrency_Level;
                Env_Val : A_String;
                Env_Len : Integer;
                Flag_Ok : Boolean;
                Os_Status : Interfaces.C.Int;


                -- PATCH: Missing from 3.2.0 predefineds.
                -- Add here to allow customers to compile this.

                use Interfaces.C;
                function Pthread_Setconcurrency (New_Level : Int) return Int;
                pragma Import (C, Pthread_Setconcurrency, "thr_setconcurrency");

            begin
                if Concurrency_Level <= 1 then
                    Env_Val := Getenv (Concurrency_Env_Variable_String'Address);
                    if Env_Val = null then
                        Env_Len := -1;
                    else
                        for I in Env_Val.all'Range loop
                            Env_Len := I - 1;
                            exit when Env_Val.all (I) = Ascii.Nul;
                        end loop;
                    end if;
                    if Env_Len > 0 then
                        A_To_I (Env_Val.all (1 .. Env_Len),
                                Concurrency_Level, Flag_Ok);
                        if Flag_Ok = False then
                            raise Ada_Concurrency_Level_Error;
                        end if;
                        Os_Status := Pthread_Setconcurrency
                                        (Interfaces.C.Int (Concurrency_Level));
                    end if;
                end if;
            end Concurrency_Init;

            Return_Value : Integer;
        begin
            Return_Value := Exception_Unwind_Safe.Call_Proc
                               (Concurrency_Init'Address);
        end;
    end V_Start_Program_Continue_Ada_2;

    procedure V_Start_Program_Continue_Ada is
        Return_Value : Integer;
    begin
        V_Start_Program_Continue_Ada_1;
        V_Start_Program_Continue_Ada_2;


        Return_Value := Exception_Unwind_Safe.Call_Proc_Table
                           (V_Usr_Data.Elaboration_Table'Address) (1);
        -- no return;
    end V_Start_Program_Continue_Ada;


    procedure V_Integrated_Init
                 -- Called before elaboration of the ada program and before
                 -- elaboration of the C++ program for VADS C++.  Do the initialization
                 -- required to support C++ exception initialization and elaboration.
                 -- Mainly this is getting the memory allocater AA_INIT initialized.
                 is
    begin
        V_Start_Program_Common (V_Start_Program_Continue_Ada_1'Address);
    end V_Integrated_Init;

    -- Called to fully initialize and elaborate the ada program.
    function V_Adainit (Argc : Integer; Argv : Address) return Integer is
        Return_Value : Integer;
    begin
        if not V_Adainit_Called then
            -- can be called multiple times but only effective first time.
            V_Adainit_Called := True;
            V_Usr_Conf_I.Argc_Value := Argc;
            V_Usr_Conf_I.Argv_Value := Argv;
            V_Start_Program_Common (V_Start_Program_Continue_Ada_1'Address);
            V_Start_Program_Continue_Ada_2;  -- do rest of initialization

            Return_Value := Exception_Unwind_Safe.Call_Proc_Table_Minus_One
                               (V_Usr_Data.Elaboration_Table'Address) (1);
            return Return_Value;
        end if;
        return -1;
    end V_Adainit;

    procedure V_Adafinal is

        procedure Ts_Final;
        pragma Interface (Ada, Ts_Final);
        pragma Interface_Name (Ts_Final, "TS_FINAL");
    begin
        if not V_Adafinal_Called then
            V_Adafinal_Called := True;
            if V_Adainit_Called then
                Ts_Final;
            end if;
        end if;
    end V_Adafinal;
    -- THE FOLLOWING ROUTINES SHOULD NOT BE MODIFIED


    procedure V_Start_Program_Continue is
        pragma Implicit_Code (Off);
        use Machine_Code;
    begin
        -- Before clearing fp, need to flush windows.
        Code_1'(Ta, +Psignal.St_Flush_Windows);

        -- Null the frame pointer to indicate top of stack for the program.
        Code_1'(Clr, Fp);

        Code_1'(Call, V_Start_Program_Continue_Ada'Ref);
        Code_0'(Op => Nop);
        -- No return, but just in case exit the user program

        Code_1'(Call, V_Usr_Conf_I.Ada_Exit'Ref);
        Code_2'(Mov, G0, O0);        -- return 0
    end V_Start_Program_Continue;

end V_Usr_Conf;
